# From http://stackoverflow.com/questions/36932/whats-the-best-way-to-implement-an-enum-in-python
def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)




__version__ = "0.1.7"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(HEAD detached at v0.1.7)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '26fa564aea92a5640ab3cf1be333324566f41896'
__git_head_subject__ = 'Finished Release v0.1.7'
__git_head_message__ = ''
__git_dirty_diff__ = ''


import ctypes
from .service import ServiceState, ServiceControlsAccepted, SERVICE_STATUS, Service
from .common import ServiceControl, ServiceType

import logging
logger = logging.getLogger(__name__)

RegisterServiceCtrlHandlerEx = ctypes.windll.advapi32.RegisterServiceCtrlHandlerExW
StartServiceCtrlDispatcher   = ctypes.windll.advapi32.StartServiceCtrlDispatcherW

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms685138%28v=VS.85%29.aspx
# VOID WINAPI ServiceMain(
#   __in  DWORD dwArgc,
#   __in  LPTSTR *lpszArgv
# );
SERVICE_MAIN_FUNCTION = ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_ulong, ctypes.POINTER(ctypes.c_wchar_p))

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms683241%28v=VS.85%29.aspx
# DWORD WINAPI HandlerEx(
#   __in  DWORD dwControl,
#   __in  DWORD dwEventType,
#   __in  LPVOID lpEventData,
#   __in  LPVOID lpContext
# );
# Although lpContext is LPVOID here, we specify c_ulong because we store the Pythonic context object too so it won't
# get garbage collected. We use Python's id() function as the "address" of the Pythonic context object.
# See the rest in register_ctrl_handler
HANDLER_EX = ctypes.WINFUNCTYPE(ctypes.c_ulong, ctypes.c_ulong, ctypes.c_ulong, ctypes.c_void_p, ctypes.c_uint)

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms686324%28v=VS.85%29.aspx
# typedef struct _SERVICE_TABLE_ENTRY {
#   LPTSTR                  lpServiceName;
#   LPSERVICE_MAIN_FUNCTION lpServiceProc;
# } SERVICE_TABLE_ENTRY, *LPSERVICE_TABLE_ENTRY;
class SERVICE_TABLE_ENTRY(ctypes.Structure):
    _fields_ = [("lpServiceName", ctypes.c_wchar_p), ("lpServiceProc", SERVICE_MAIN_FUNCTION)]

class _ServiceCtrl(object):
    def __init__(self):
        self._garbage_protect_map = dict()
        self._contexts = dict()
        self._handles = dict()

    def register_ctrl_handler(self, service_name, callback, context=None):
        def wrapper(dwControl, dwEventType, lpEventData, lpContext):
            try:
                if dwControl == ServiceControl.DEVICEEVENT and dwEventType == EventType.POWERSETTINGCHANGE:
                    # TODO: convert lpEventData to a POWERBROADCAST_SETTING structure.
                    pass
                elif dwControl == ServiceControl.SESSIONCHANGE:
                    # TODO: convert lpEventData to WTSSESSION_NOTIFICATION structure.
                    pass
                elif dwControl == ServiceControl.TIMECHANGE:
                    # TODO: convert lpEventData to SERVICE_TIMECHANGE_INFO structure.
                    pass

                if id(wrapper) in self._handles:
                    handle = self._handles[id(wrapper)]
                else:
                    handle = None
                if lpContext in self._contexts:
                    context = self._contexts[lpContext]
                else:
                    context = None
                    
                return callback(handle, dwControl, dwEventType, lpEventData, context)
            except:
                logger.exception("exception caught in service callback handler")
            
        thunk = HANDLER_EX(wrapper)

        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms685058%28v=VS.85%29.aspx
        # SERVICE_STATUS_HANDLE WINAPI RegisterServiceCtrlHandlerEx(
        #   __in      LPCTSTR lpServiceName,
        #   __in      LPHANDLER_FUNCTION_EX lpHandlerProc,
        #   __in_opt  LPVOID lpContext
        # );
        handle = RegisterServiceCtrlHandlerEx(unicode(service_name), thunk, id(context))
        if handle == 0:
            raise ctypes.WinError()

        self._garbage_protect_map[id(thunk)] = thunk
        self._contexts[id(context)] = context
        self._handles[id(wrapper)] = handle
            
        return Service(handle)

    def start_ctrl_dispatcher(self, *services):
        """
        Sets the ServiceMain function of each service. Every element in the list is a pair of name and callback.
        
        For example:
        >>> start_service_ctrl_dispatcher(("my_service", my_service_main), ("my_other_service", my_other_service_main))

        From the MSDN docs, if the service is SERVICE_WIN32_OWN_PROCESS then the service name is ignored
        (but cannot be NULL).
        """
        # We create a NULL terminated array of pointers to SERVICE_TABLE_ENTRYs.
        service_tables = (SERVICE_TABLE_ENTRY * (len(services) + 1))()
        for i, service in enumerate(services):
            # We wrap the normal ServiceMain so we can pass the Python callback a nice argv Python array.
            def main_wrapper(argc, argv):
                try:
                    service[1](list(argv[i] for i in xrange(argc)))
                except:
                    logger.exception("service main exception caught")

            thunk = SERVICE_MAIN_FUNCTION(main_wrapper)
            name = unicode(service[0]) if service[0] is not None else u""
            service_tables[i] = SERVICE_TABLE_ENTRY(lpServiceName=name, lpServiceProc=thunk)

        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms686324%28v=VS.85%29.aspx
        # BOOL WINAPI StartServiceCtrlDispatcher(
        #   __in  const SERVICE_TABLE_ENTRY *lpServiceTable
        # );
        if not StartServiceCtrlDispatcher(service_tables):
            raise ctypes.WinError()

        # We need to protect ctype's thunk from the garbage collector. Note that this is a memory "leak" because we
        # never remove it from the map, but it's okay since a service should have a single ServiceMain entry.
        for service_table in service_tables:
            self._garbage_protect_map[id(service_table.lpServiceProc)] = service_table.lpServiceProc

ServiceCtrl = _ServiceCtrl()

class ServiceRunner(object):
    def __init__(self, service_name):
        self.status = ServiceState.START_PENDING
        self.service_name = service_name
        
    def main(self):
        raise NotImplementedError()
    
    def control(self, service_control):
        raise NotImplementedError()
    
    def run(self):
        logger.debug("ServiceRunner.run called.")
        try:
            ServiceCtrl.start_ctrl_dispatcher((self.service_name, self._service_main))
        except:
            logger.exception("error occurred")
            
    def _service_main(self, args):
        logger.debug("ServiceRunner._service_main called, self=%s, args=%s" % (self, repr(args)))
        
        try:
            service = ServiceCtrl.register_ctrl_handler(self.service_name, self._service_callback)
            
            logger.debug("setting status to START_PENDING")
            self._notify_status(service, ServiceState.START_PENDING)
        
            logger.debug("setting status to RUNNING")
            self._notify_status(service, ServiceState.RUNNING)
        
            self.main()
        except:
            logger.exception("error occurred")

    def _service_callback(self, handle, fdwControl, dwEventType, lpEventData, lpContext):
        logger.debug("ServiceRunner._service_callback: handle=%x, fdwControl=%x dwEventType=%x" %
                  (handle, fdwControl, dwEventType))

        self.control(fdwControl)

        service = Service(handle)
        if fdwControl == ServiceControl.STOP:
            logger.debug("STOP requested, quitting.")
            self._notify_status(service, ServiceState.STOPPED)
        elif fdwControl == ServiceControl.INTERROGATE:
            logger.debug("INTERROGATE requested.")
            self._notify_status(service)
                    
        return 0

    def _notify_status(self, service, status=None):
        if status is not None:
            self.status = status
        status_struct = SERVICE_STATUS(dwServiceType=ServiceType.WIN32_OWN_PROCESS,
                                       dwCurrentState=self.status,
                                       dwControlsAccepted=(ServiceControlsAccepted.STOP | 
                                                           ServiceControlsAccepted.SHUTDOWN),
                                       dwWin32ExitCode=0,
                                       dwServiceSpecificExitCode=0,
                                       dwCheckPoint=0,
                                       dwWaitHint=0)
        service.set_status(status_struct)



import ctypes
from collections import namedtuple

from .utils import enum
from .service import Service
from .common import ServiceType, ERROR_INVALID_HANDLE

OpenSCManager      = ctypes.windll.advapi32.OpenSCManagerW
OpenService        = ctypes.windll.advapi32.OpenServiceW
CloseServiceHandle = ctypes.windll.advapi32.CloseServiceHandle
CreateService      = ctypes.windll.advapi32.CreateServiceW

# From http://msdn.microsoft.com/en-us/library/windows/desktop/ms685981%28v=vs.85%29.aspx
ServiceManagerAccess = enum(
    ALL                = 0xF003F,
    CONNECT            = 0x0001,
    CREATE_SERVICE     = 0x0002,
    ENUMERATE_SERVICE  = 0x0004,
    LOCK               = 0x0008,
    QUERY_LOCK_STATUS  = 0x0010,
    MODIFY_BOOT_CONFIG = 0x0020)

# From http://msdn.microsoft.com/en-us/library/windows/desktop/ms684323%28v=vs.85%29.aspx
SC_ACTIVE_DATABASE = u"SERVICES_ACTIVE_DATABASE"

# From http://msdn.microsoft.com/en-us/library/windows/desktop/ms682450%28v=vs.85%29.aspx

# -- CreateService.dwStartType:
ServiceStartType = enum(
    BOOT     = 0x00000000,
    SYSTEM   = 0x00000001,
    AUTO     = 0x00000002,
    DEMAND   = 0x00000003,
    DISABLED = 0x00000004)

# -- CreateService.dwErrorControl:
ServiceErrorControl = enum(
    IGNORE   = 0x00000000, 
    NORMAL   = 0x00000001,
    SEVERE   = 0x00000002,
    CRITICAL = 0x00000003)

# -- CreateService.dwDesiredAccess:
ServiceAccess = enum(ALL                  = 0xF01FF,
                     QUERY_CONFIG         = 0x0001,
                     CHANGE_CONFIG        = 0x0002,
                     QUERY_STATUS         = 0x0004,
                     ENUMERATE_DEPENDENTS = 0x0008,
                     START                = 0x0010,
                     STOP                 = 0x0020,
                     PAUSE_CONTINUE       = 0x0040,
                     INTERROGATE          = 0x0080,
                     USER_DEFINED_CONTROL = 0x0100)

class ServiceControlManagerContext(object):
    def __init__(self, machine=None, database=None, access=ServiceManagerAccess.ALL):
        super(ServiceControlManagerContext, self).__init__()
        self.machine = unicode(machine) if machine is not None else 0
        self.database = unicode(database) if machine is not None else 0
        self.access = access
        self.scm = None
        
    def __enter__(self):
        scm_handle = OpenSCManager(self.machine, self.database, self.access)
        if scm_handle == 0:
            raise ctypes.WinError()
        self.scm = ServiceControlManager(scm_handle)
        return self.scm

    def __exit__(self, type, value, traceback):
        if self.scm is not None:
            self.scm.close()

class ServiceControlManager(object):
    def __init__(self, handle):
        super(ServiceControlManager, self).__init__()
        self.handle = ctypes.c_void_p(handle) if isinstance(handle, (int, long)) else \
                      ctypes.c_void_p(handle.value) if hasattr(handle, value) else handle

    def create_service(self, name, display_name, type, start_type, path,
                       load_order_group=None, dependencies=None, error_control=ServiceErrorControl.NORMAL,
                       account=None, account_password=None, access=ServiceAccess.ALL):
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms682450%28v=vs.85%29.aspx
        # SC_HANDLE WINAPI CreateService(
        #   __in       SC_HANDLE hSCManager,
        #   __in       LPCTSTR lpServiceName,
        #   __in_opt   LPCTSTR lpDisplayName,
        #   __in       DWORD dwDesiredAccess,
        #   __in       DWORD dwServiceType,
        #   __in       DWORD dwStartType,
        #   __in       DWORD dwErrorControl,
        #   __in_opt   LPCTSTR lpBinaryPathName,
        #   __in_opt   LPCTSTR lpLoadOrderGroup,
        #   __out_opt  LPDWORD lpdwTagId,
        #   __in_opt   LPCTSTR lpDependencies,
        #   __in_opt   LPCTSTR lpServiceStartName,
        #   __in_opt   LPCTSTR lpPassword
        #);
        lpServiceName = unicode(name)
        lpDisplayName = unicode(display_name)
        dwDesiredAccess = access
        dwServiceType = type
        dwStartType = start_type
        dwErrorControl = error_control
        lpBinaryPathName = unicode(path) 
        lpLoadOrderGroup = unicode(load_order_group) if load_order_group is not None else None

        # TODO: Setting a lpdwTagId makes Windows return error 87 "The parameter is incorrect". Since it's not that
        # important at the moment, we'll keep this as NULL.
        tag_id = ctypes.c_uint(0)
        # lpdwTagId = ctypes.addressof(tag_id)
        lpdwTagId = None

        lpDependencies = unicode(dependencies) if dependencies is not None else None
        lpServiceStartName = unicode(account) if account is not None else None
        lpPassword = unicode(account_password) if account_password is not None else None

        assert self.handle != 0
        service_h = CreateService(self.handle, lpServiceName, lpDisplayName, dwDesiredAccess, dwServiceType,
                                  dwStartType, dwErrorControl, lpBinaryPathName, lpLoadOrderGroup, lpdwTagId,
                                  lpDependencies, lpServiceStartName, lpPassword)
        if service_h == 0:
            raise ctypes.WinError()
        return Service(service_h, tag_id.value)

    def open_service(self, name, access=ServiceAccess.ALL):
        service_h = OpenService(self.handle, unicode(name), access)
        if service_h == 0:
            raise ctypes.WinError()
        return Service(service_h)

    def close(self):
        if self.handle != 0:
            if not CloseServiceHandle(self.handle):
                if ctypes.GetLastError() != ERROR_INVALID_HANDLE:
                    raise ctypes.WinError()
            self.handle = 0

    def is_service_exist(self, name):
        try:
            service = self.open_service(name, ServiceAccess.QUERY_STATUS)
            service.close()
            return True
        except WindowsError:
            return False

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



__import__("pkg_resources").declare_namespace(__name__)

from .utils import enum

from .service import SERVICE_STATUS, SERVICE_NOTIFY, ServiceState, ServiceControlsAccepted, Service
from .common import *
from .service_runner import ServiceCtrl, ServiceRunner

from .service_control_manager import ServiceManagerAccess, SC_ACTIVE_DATABASE, ServiceStartType
from .service_control_manager import ServiceErrorControl, ServiceAccess
from .service_control_manager import ServiceControlManagerContext, ServiceControlManager




from .utils import enum

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms683241%28v=VS.85%29.aspx
ServiceControl = enum(
    STOP                  = 0x00000001,
    PAUSE                 = 0x00000002,
    CONTINUE              = 0x00000003,
    INTERROGATE           = 0x00000004,
    SHUTDOWN              = 0x00000005,
    PARAMCHANGE           = 0x00000006,
    NETBINDADD            = 0x00000007,
    NETBINDREMOVE         = 0x00000008,
    NETBINDENABLE         = 0x00000009,
    NETBINDDISABLE        = 0x0000000A,
    DEVICEEVENT           = 0x0000000B,
    HARDWAREPROFILECHANGE = 0x0000000C,
    POWEREVENT            = 0x0000000D,
    SESSIONCHANGE         = 0x0000000E,
    PRESHUTDOWN           = 0x0000000F,
    TIMECHANGE            = 0x00000010,
    TRIGGEREVENT          = 0x00000020
    # 0x80 - 0xFF: user defined
    )

# From http://msdn.microsoft.com/en-us/library/windows/desktop/ms682450%28v=vs.85%29.aspx
# Also, from http://msdn.microsoft.com/en-us/library/windows/desktop/ms685996%28v=vs.85%29.aspx
# -- CreateService.dwServiceType and SERVICE_STATUS structure used in SetServiceStatus (but only a subset is available
#    there):
ServiceType = enum(
    KERNEL_DRIVER         = 0x00000001,
    FILE_SYSTEM_DRIVER    = 0x00000002,
    ADAPTER               = 0x00000004,
    RECOGNIZER_DRIVER     = 0x00000008,
    WIN32_OWN_PROCESS     = 0x00000010,
    WIN32_SHARE_PROCESS   = 0x00000020,
    INTERACTIVE_PROCESS   = 0x00000100)

ERROR_INVALID_HANDLE = 6



import ctypes
import logging

from .utils import enum
from .common import ServiceControl, ServiceType, ERROR_INVALID_HANDLE

StartService                 = ctypes.windll.advapi32.StartServiceW
ControlService               = ctypes.windll.advapi32.ControlService
DeleteService                = ctypes.windll.advapi32.DeleteService
SetServiceStatus             = ctypes.windll.advapi32.SetServiceStatus
CloseServiceHandle           = ctypes.windll.advapi32.CloseServiceHandle
QueryServiceStatus           = ctypes.windll.advapi32.QueryServiceStatus
QueryServiceConfig           = ctypes.windll.advapi32.QueryServiceConfigW
ChangeServiceConfig          = ctypes.windll.advapi32.ChangeServiceConfigW

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms685992%28v=VS.85%29.aspx
# typedef struct _SERVICE_STATUS_PROCESS {
#   DWORD dwServiceType;
#   DWORD dwCurrentState;
#   DWORD dwControlsAccepted;
#   DWORD dwWin32ExitCode;
#   DWORD dwServiceSpecificExitCode;
#   DWORD dwCheckPoint;
#   DWORD dwWaitHint;
#   DWORD dwProcessId;
#   DWORD dwServiceFlags;
# } SERVICE_STATUS_PROCESS, *LPSERVICE_STATUS_PROCESS;
class SERVICE_STATUS_PROCESS(ctypes.Structure):
    _fields_ = [("dwServiceType", ctypes.c_ulong),
                ("dwCurrentState", ctypes.c_ulong),
                ("dwControlsAccepted", ctypes.c_ulong),
                ("dwWin32ExitCode", ctypes.c_ulong),
                ("dwServiceSpecificExitCode", ctypes.c_ulong),
                ("dwCheckPoint", ctypes.c_ulong),
                ("dwWaitHint", ctypes.c_ulong),
                ("dwProcessId", ctypes.c_ulong),
                ("dwServiceFlags", ctypes.c_ulong)]

# https://msdn.microsoft.com/en-us/library/windows/desktop/ms684950(v=vs.85).aspx
# typedef struct _QUERY_SERVICE_CONFIG {
#   DWORD  dwServiceType;
#   DWORD  dwStartType;
#   DWORD  dwErrorControl;
#   LPTSTR lpBinaryPathName;
#   LPTSTR lpLoadOrderGroup;
#   DWORD  dwTagId;
#   LPTSTR lpDependencies;
#   LPTSTR lpServiceStartName;
#   LPTSTR lpDisplayName;
# } QUERY_SERVICE_CONFIG, *LPQUERY_SERVICE_CONFIG;
class QUERY_SERVICE_CONFIG(ctypes.Structure):
    _fields_ = [("dwServiceType", ctypes.c_ulong),
                ("dwStartType", ctypes.c_ulong),
                ("dwErrorControl", ctypes.c_ulong),
                ("lpBinaryPathName", ctypes.c_wchar_p),
                ("lpLoadOrderGroup", ctypes.c_wchar_p),
                ("dwTagId", ctypes.c_ulong),
                ("lpDependencies", ctypes.c_wchar_p),
                ("lpServiceStartName", ctypes.c_wchar_p)]

    def to_dict(self):
        return dict(service_type=self.dwServiceType, start_type=self.dwStartType,
                    error_control=self.dwErrorControl, binary_path_name=self.lpBinaryPathName,
                    load_order_group=self.lpLoadOrderGroup, tag_id=self.dwTagId,
                    dependencies=self.lpDependencies, service_start_name=self.lpServiceStartName)


LPQUERY_SERVICE_CONFIG = ctypes.POINTER(QUERY_SERVICE_CONFIG)

# From http://msdn.microsoft.com/en-us/library/windows/desktop/ms685996%28v=vs.85%29.aspx

ServiceState = enum(
    STOPPED          = 0x00000001,
    START_PENDING    = 0x00000002,
    STOP_PENDING     = 0x00000003,
    RUNNING          = 0x00000004,
    CONTINUE_PENDING = 0x00000005,
    PAUSE_PENDING    = 0x00000006,
    PAUSED           = 0x00000007
)

ServiceControlsAccepted = enum(
    STOP                  = 0x00000001,
    SHUTDOWN              = 0x00000004,
    PARAMCHANGE           = 0x00000008,
    PAUSE_CONTINUE        = 0x00000002,
    NETBINDCHANGE         = 0x00000010,
    HARDWAREPROFILECHANGE = 0x00000020,
    POWEREVENT            = 0x00000040,
    SESSIONCHANGE         = 0x00000080,
    PRESHUTDOWN           = 0x00000100,
    TIMECHANGE            = 0x00000200,
    TRIGGEREVENT          = 0x00000400
)


ServiceType = enum(
    SERVICE_KERNEL_DRIVER = 0x00000001,
    SERVICE_FILE_SYSTEM_DRIVER = 0x00000002,
    SERVICE_WIN32_OWN_PROCESS = 0x00000010,
    SERVICE_WIN32_SHARE_PROCESS = 0x00000020
)

StartType = enum(
SERVICE_BOOT_START = 0x00000000,
SERVICE_SYSTEM_START = 0x00000001,
SERVICE_AUTO_START = 0x00000002,
SERVICE_DEMAND_START = 0x00000003,
SERVICE_DISABLED = 0x00000004
)

# typedef struct _SERVICE_STATUS {
#   DWORD dwServiceType;
#   DWORD dwCurrentState;
#   DWORD dwControlsAccepted;
#   DWORD dwWin32ExitCode;
#   DWORD dwServiceSpecificExitCode;
#   DWORD dwCheckPoint;
#   DWORD dwWaitHint;
# } SERVICE_STATUS, *LPSERVICE_STATUS;
class SERVICE_STATUS(ctypes.Structure):
    _fields_ = [("dwServiceType", ctypes.c_ulong),
                ("dwCurrentState", ctypes.c_ulong),
                ("dwControlsAccepted", ctypes.c_ulong),
                ("dwWin32ExitCode", ctypes.c_ulong),
                ("dwServiceSpecificExitCode", ctypes.c_ulong),
                ("dwCheckPoint", ctypes.c_ulong),
                ("dwWaitHint", ctypes.c_ulong)]

LPSERVICE_STATUS = ctypes.POINTER(SERVICE_STATUS)

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms685947%28v=VS.85%29.aspx
# typedef VOID( CALLBACK * PFN_SC_NOTIFY_CALLBACK ) (
#     IN PVOID pParameter
# );
FN_SC_NOTIFY_CALLBACK = ctypes.WINFUNCTYPE(ctypes.c_int, ctypes.c_void_p)

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms685947%28v=VS.85%29.aspx
# typedef struct _SERVICE_NOTIFY {
#   DWORD                  dwVersion;
#   PFN_SC_NOTIFY_CALLBACK pfnNotifyCallback;
#   PVOID                  pContext;
#   DWORD                  dwNotificationStatus;
#   SERVICE_STATUS_PROCESS ServiceStatus;
#   DWORD                  dwNotificationTriggered;
#   LPTSTR                 pszServiceNames;
# } SERVICE_NOTIFY, *PSERVICE_NOTIFY;
class SERVICE_NOTIFY(ctypes.Structure):
    _fields_ = [("dwVersion", ctypes.c_ulong),
                ("pfnNotifyCallback", FN_SC_NOTIFY_CALLBACK),
                ("pContext", ctypes.c_void_p),
                ("dwNotificationStatus", ctypes.c_ulong),
                ("ServiceStatus", SERVICE_STATUS_PROCESS),
                ("dwNotificationTriggered", ctypes.c_ulong),
                ("pszServiceNames", ctypes.c_wchar_p)]

# http://msdn.microsoft.com/en-us/library/windows/desktop/ms684276%28v=VS.85%29.aspx
ServiceNotifyMask = enum(
    STOPPED          = 0x00000001,
    START_PENDING    = 0x00000002,
    STOP_PENDING     = 0x00000004,
    RUNNING          = 0x00000008,
    CONTINUE_PENDING = 0x00000010,
    PAUSE_PENDING    = 0x00000020,
    PAUSED           = 0x00000040,
    CREATED          = 0x00000080,
    DELETED          = 0x00000100,
    DELETE_PENDING   = 0x00000200
)

# From WinError.h:
ERROR_SERVICE_SPECIFIC_ERROR = 1066
NO_ERROR = 0

# From winsvc.h:
SERVICE_NO_CHANGE = 0xffffffff

class Service(object):
    def __init__(self, handle, tag=None):
        self.handle = ctypes.c_void_p(handle) if isinstance(handle, (int, long)) else \
                      ctypes.c_void_p(handle.value) if hasattr(handle, value) else handle
        self.tag = tag

    def start(self, *args):
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms686321%28v=vs.85%29.aspx
        # BOOL WINAPI StartService(
        #   __in      SC_HANDLE hService,
        #   __in      DWORD dwNumServiceArgs,
        #   __in_opt  LPCTSTR *lpServiceArgVectors
        # );
        if len(args) == 0:
            lpServiceArgVectors = None
        else:
            lpServiceArgVectors = (ctypes.c_wchar_p * len(args))(*args)
        if not StartService(self.handle, len(args), lpServiceArgVectors):
            raise ctypes.WinError()

    def wait_on_pending(self, timeout_in_seconds=60):
        from time import sleep
        for sec in xrange(timeout_in_seconds):
            if self.get_status() in (ServiceState.STOP_PENDING, ServiceState.START_PENDING):
                sleep(1)
            else:
                return
        raise RuntimeError("wait_on_pending timed out, status is: {}".format(self.get_status()))

    def stop(self):
        """
        Stops the service.
        """
        new_status = SERVICE_STATUS()
        if not ControlService(self.handle, ServiceControl.STOP, ctypes.byref(new_status)):
            raise ctypes.WinError()
        if new_status.dwCurrentState not in [ServiceState.STOPPED, ServiceState.STOP_PENDING]:
            raise ctypes.WinError()

    def safe_start(self):
        if self.get_status() in [ServiceState.RUNNING, ServiceState.START_PENDING]:
            return
        self.start()

    def safe_stop(self):
        """
        Stops the service, and ignores "not started" errors
        """
        if self.get_status() in [ServiceState.STOPPED, ServiceState.STOP_PENDING]:
            return
        try:
            self.stop()
        except WindowsError, e:
            if e.winerror != 1062:
                raise

    def get_status(self):
        current_status = SERVICE_STATUS()
        if not QueryServiceStatus(self.handle, ctypes.byref(current_status)):
            raise ctypes.WinError()
        return current_status.dwCurrentState

    def is_running(self):
        return self.get_status() == ServiceState.RUNNING

    # def change_service_config
    # https://msdn.microsoft.com/en-us/library/windows/desktop/ms681987(v=vs.85).aspx

    def query_config(self):
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms684932%28v=vs.85%29.aspx
        # BOOL WINAPI QueryServiceConfig(
        #   __in       SC_HANDLE hService,
        #   __out_opt  LPQUERY_SERVICE_CONFIG lpServiceConfig,
        #   __in       DWORD cbBufSize,
        #   __out      LPDWORD pcbBytesNeeded
        # );
        config_buffer = ctypes.create_string_buffer(8192) # The maximum size of this array is 8K bytes
        bytes_needed = ctypes.c_ulong()
        if not QueryServiceConfig(self.handle, config_buffer, 8192, ctypes.byref(bytes_needed)):
            raise ctypes.WinError()
        service_config = ctypes.cast(config_buffer, ctypes.POINTER(QUERY_SERVICE_CONFIG))
        return service_config.contents.to_dict()

    def change_service_config(self, start_type):
        # https://msdn.microsoft.com/en-us/library/windows/desktop/ms681987(v=vs.85).aspx
        # BOOL WINAPI ChangeServiceConfig(
        #   _In_      SC_HANDLE hService,
        #   _In_      DWORD     dwServiceType,
        #   _In_      DWORD     dwStartType,
        #   _In_      DWORD     dwErrorControl,
        #   _In_opt_  LPCTSTR   lpBinaryPathName,
        #   _In_opt_  LPCTSTR   lpLoadOrderGroup,
        #   _Out_opt_ LPDWORD   lpdwTagId,
        #   _In_opt_  LPCTSTR   lpDependencies,
        #   _In_opt_  LPCTSTR   lpServiceStartName,
        #   _In_opt_  LPCTSTR   lpPassword,
        #   _In_opt_  LPCTSTR   lpDisplayName
        # );
        null_ptr = ctypes.POINTER(ctypes.c_int)()
        if not ChangeServiceConfig(self.handle,
                                   SERVICE_NO_CHANGE, start_type, SERVICE_NO_CHANGE,
                                   null_ptr, null_ptr,
                                   null_ptr, null_ptr, null_ptr,
                                   null_ptr, null_ptr):
            raise ctypes.WinError()

    def is_disabled(self):
        return self.query_config()['start_type'] == StartType.SERVICE_DISABLED

    def is_autostart(self):
        return self.query_config()['start_type'] == StartType.SERVICE_AUTO_START

    def disable(self):
        self.change_service_config(StartType.SERVICE_DISABLED)

    def start_automatically(self):
        self.change_service_config(StartType.SERVICE_AUTO_START)

    def query_optional_config(self):
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms684935%28v=VS.85%29.aspx
        # BOOL WINAPI QueryServiceConfig2(
        #   __in       SC_HANDLE hService,
        #   __in       DWORD dwInfoLevel,
        #   __out_opt  LPBYTE lpBuffer,
        #   __in       DWORD cbBufSize,
        #   __out      LPDWORD pcbBytesNeeded
        # );
        raise NotImplementedError()

    def set_status(self, status):
        """
        Sets the service status. status argument must be a SERVICE_STATUS object.
        """
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms686241%28v=VS.85%29.aspx
        # BOOL WINAPI SetServiceStatus(
        #   __in  SERVICE_STATUS_HANDLE hServiceStatus,
        #   __in  LPSERVICE_STATUS lpServiceStatus
        # );
        if not SetServiceStatus(self.handle, LPSERVICE_STATUS(status)):
            raise ctypes.WinError()

    def delete(self):
        """
        Deletes the service.
        """
        # http://msdn.microsoft.com/en-us/library/windows/desktop/ms682562%28v=vs.85%29.aspx
        # BOOL WINAPI DeleteService(
        #   __in  SC_HANDLE hService
        # );
        if not DeleteService(self.handle):
            raise ctypes.WinError()

    def close(self):
        if self.handle != 0:
            if not CloseServiceHandle(self.handle):
                if ctypes.GetLastError() != ERROR_INVALID_HANDLE:
                    raise ctypes.WinError()
            self.handle = 0

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
